/* ----------------------------------------------------------------------- *
 *   
 *   Copyright 2004-2009 H. Peter Anvin - All Rights Reserved
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, Inc., 53 Temple Place Ste 330,
 *   Bostom MA 02111-1307, USA; either version 2 of the License, or
 *   (at your option) any later version; incorporated herein by reference.
 *
 * ----------------------------------------------------------------------- */

/*
 * abcprintd.c
 *
 * Read a serial port; send data to lpr when receiving FF 00
 *
 * Usage: abcprintd port [options to lpr]
 */

#include <stdio.h>
#include <unistd.h>
#include <termios.h>
#include <stdlib.h>
#include <string.h>
#include <stdbool.h>
#include <sys/wait.h>
#include <fcntl.h>

#ifndef O_TEXT
# define O_TEXT 0
#endif
#ifndef O_BINARY
# define O_BINARY 0
#endif

static int lpr_argc;
static const char **lpr_argv;

enum print_state {
  ps_first,			/* Brand new job */
  ps_percent,			/* Job started with %, might be ps */
  ps_binary,			/* Don't touch this */
  ps_text			/* It's text, OK to modify */
};

static void print_setup(FILE **tfp, enum print_state *psp)
{
  FILE *tf = *tfp;
  pid_t f;

  if ( tf ) {
    fflush(tf);
    rewind(tf);

    f = fork();

    if ( f < 0 ) {
      perror("fork");
      exit(1);
    } else if ( f == 0 ) {
      dup2(fileno(tf), STDIN_FILENO);
      fclose(tf);
      execvp(lpr_argv[0], (char **)lpr_argv);
      _exit(255);
    } else {
      fclose(tf);
      while ( waitpid(f, NULL, 0) != f );
    }
  }

  *tfp = tmpfile();
  if ( !*tfp ) {
    perror("tmpfile");
    exit(1);
  }
  *psp = ps_first;
}

static int open_port(const char *path, speed_t speed)
{
  struct termios tio;
  int fd;
  int flags;

  if ( (fd = open(path, O_RDWR|O_NONBLOCK|O_BINARY)) < 0 )
    return -1;

  if ( tcgetattr(fd, &tio) )
    return -1;

  tio.c_iflag &= ~(BRKINT|PARMRK|INPCK|ISTRIP|ICRNL|INLCR|IGNCR|
		   IXON|IXOFF);
  tio.c_iflag |= IGNBRK|IGNPAR;

  tio.c_cflag &= ~(CSIZE|CSTOPB|HUPCL|PARENB|PARODD);
  tio.c_cflag |= CLOCAL|CREAD|CS8;

  tio.c_oflag &= ~OPOST;

  tio.c_lflag &= ~(ECHO|ECHOE|ECHOK|ECHONL|ICANON|ISIG|NOFLSH|TOSTOP|IEXTEN);

  tio.c_cc[VMIN]  = 1;
  tio.c_cc[VTIME] = 0;

  
  cfsetispeed(&tio, speed);
  cfsetospeed(&tio, speed);

  tcflush(fd, TCIFLUSH);

  if ( tcsetattr(fd, TCSANOW, &tio) )
    return -1;

  if ( (flags = fcntl(fd, F_GETFL)) == -1 )
    return -1;

  flags &= ~O_NONBLOCK;

  if ( fcntl(fd, F_SETFL, flags) )
    return -1;

  return fd;
}

static void output(int c, FILE *tf, enum print_state *psp)
{
  static const char abc_to_latin1[256] =
    "\000\001\002\003\004\005\006\007\010\011\012\013\014\015\016\017"
    "\020\021\022\023\024\025\026\027\030\031\032\033\034\035\036\037"
    " !\"#\244%&\'()*+,-./0123456789:;<=>?"
    "\311ABCDEFGHIJKLMNOPQRSTUVWXYZ\304\326\305\334_"
    "\351abcdefghijklmnopqrstuvwxyz\344\366\345\374\267"
    "\200\201\202\203\204\205\206\207\210\211\212\213\214\215\216\217"
    "\220\221\222\223\224\225\226\227\230\231\232\233\234\235\236\237"
    "\240\241\242\243$\245\246\247\250\251\252\253\254\255\256\257"
    "\260\261\262\263\264\265\266\177\270\271\272\273\274\275\276\277"
    "\300\301\302\303[]\306\307\310@\312\313\314\315\316\317"
    "\320\321\322\323\324\325\\\327\330\331\332\333^\335\336\337"
    "\340\341\342\343{}\346\347\350`\352\353\354\355\356\357"
    "\360\361\362\363\364\365|\367\370\371\372\373~\375\376\377";

  switch ( *psp ) {
  case ps_first:
    if ( c == 27 )
      *psp = ps_binary;
    else if ( c == '%' ) {
      *psp = ps_percent;
      return;
    } else {
      *psp = ps_text;
      // fwrite(text_prefix, 1, sizeof text_prefix - 1, tf);
    }
    output(c, tf, psp);
    break;

  case ps_percent:
    if ( c == '!' ) {
      *psp = ps_binary;
    } else {
      *psp = ps_text;
      // fwrite(text_prefix, 1, sizeof text_prefix - 1, tf);
    }
    output('%', tf, psp);
    output(c, tf, psp);
    break;

  case ps_text:
    putc(abc_to_latin1[(unsigned char)c], tf);
    break;

  case ps_binary:
    putc(c, tf);
    break;
  }
}

extern bool file_op(unsigned char, int);
extern const char *file_op_dir;

#define BUF_SIZE 4096

int main(int argc, char *argv[])
{
  int fd;
  FILE *tf = NULL;
  enum {
    st_normal,			/* Normal operation */
    st_ff,			/* 0xFF received */
    st_file,			/* File operation in progress */
  } state;
  char ibuf[BUF_SIZE];
  unsigned char c;
  int i, b;
  enum print_state ps;
  static const char *lpr_default_argv[] = {
    "enscript", "-B", "-Z", "-q", "-X", "latin1",
    "--non-printable-format=space",
    NULL
  };
  int o;
  extern char *optarg;
  extern int optind, opterr, optopt;
  bool badopt = false;
  speed_t speed = B115200;

  file_op_dir = NULL;

  while ((o = getopt(argc, argv, "d:t")) != EOF) {
    switch (o) {
    case 'd':
      file_op_dir = optarg;
      break;
    default:
      badopt = true;
      break;
    }
  }

  if ( badopt || optind >= argc ) {
    fprintf(stderr,
	    "Usage: %s [options] port lpr_program [lpr_args...]]\n"
	    "Options:\n"
	    "    -d dir     enable file access to directory dir\n"
	    ,
	    argv[0]);
    exit(1);
  }

  if ( (fd = open_port(argv[optind], speed)) < 0 ) {
    perror(argv[optind]);
    exit(1);
  }

  if ( argc == optind ) {
    lpr_argc = sizeof(lpr_default_argv)/sizeof(const char *) - 1;
    lpr_argv = lpr_default_argv;
  } else {
    lpr_argc = argc-optind-1;
    if ( !(lpr_argv = malloc(sizeof(char *)*(lpr_argc+1))) ) {
      perror("malloc");
    exit(1);
    }
    memcpy(lpr_argv, argv+optind+1, sizeof(char *)*(lpr_argc+1));
  }

  daemon(0,0);

  print_setup(&tf, &ps);

  state = st_normal;

  for ( ;; ) {
    if ( (b = read(fd, ibuf, sizeof ibuf)) < 1 ) {
      perror(argv[1]);
      exit(1);
    }

    for ( i = 0 ; i < b ; i++ ) {
      c = ibuf[i];

      switch ( state ) {
      case st_normal:
	if ( c == 0xff )
	  state = st_ff;
	else
	  output(c, tf, &ps);
	break;

      case st_ff:
	if ( c == 0 ) {
	  print_setup(&tf, &ps); /* BREAK received, end of job */
	  state = st_normal;
	} else if ( c >= 0xa0 && c <= 0xbf ) {
	  /* Opcode range reserved for file ops */
	  state = file_op(c, fd) ? st_file : st_normal;
	} else {
	  output(c, tf, &ps);
	  state = st_normal;
	}
	break;

      case st_file:
	state = file_op(c, fd) ? st_file : st_normal;
	break;
      }
    }
  }
}
